//
//  OpenGL widget
//
#include "ex15opengl.h"
#include <QtOpenGL>
#include <QMessageBox>
#include <math.h>
#define Cos(th) cos(M_PI/180*(th))
#define Sin(th) sin(M_PI/180*(th))

//
//  Draw vertex in polar coordinates
//
static void Vertex(double th,double ph)
{
   glTexCoord2f(th/360,ph/180+0.5);
   glNormal3f(Cos(th)*Cos(ph),Sin(th)*Cos(ph),Sin(ph));
   glVertex3d(Cos(th)*Cos(ph),Sin(th)*Cos(ph),Sin(ph));
}

//
//  Draw a ball at (x,y,z) radius r
//
static void ball(double x,double y,double z,double r)
{
   //  Save transformation
   glPushMatrix();
   //  Offset, scale and rotate
   glTranslated(x,y,z);
   glScaled(r,r,r);
   //  Bands of latitude
   for (int ph=-90;ph<90;ph+=5)
   {
      glBegin(GL_QUAD_STRIP);
      for (int th=0;th<=360;th+=10)
      {
         Vertex(th,ph);
         Vertex(th,ph+5);
      }
      glEnd();
   }
   //  Undo transofrmations
   glPopMatrix();
}

//
//  Constructor
//
Ex15opengl::Ex15opengl(QWidget* parent)
    : QGLWidget(parent)
{
   mode = 0;
   init  = false;
   mouse = false;
   asp = 1;
   dim = 1.2;
   th = ph = 0;
   pct = 0;
}

//
//  Reset view
//
void Ex15opengl::reset()
{
   th = ph = 0;
   dim = 1.2;
   //  Request redisplay
   updateGL();
}

//
//  Set shader
//
void Ex15opengl::setShader(int sel)
{
   mode = sel;
   //  Request redisplay
   updateGL();
}

//
//  Set position
//
void Ex15opengl::setPos(int p)
{
   pct = 0.01*p;
   //  Request redisplay
   updateGL();
}

//
//  Load image to texture unit
//
unsigned int Ex15opengl::LoadImage(const QString file)
{
   //  Load image
   QImage img(file);
   //  Bind texture
   unsigned int tex;
   glGenTextures(1,&tex);
   glBindTexture(GL_TEXTURE_2D,tex);
   //  Copy image to texture
   QImage rgba = QGLWidget::convertToGLFormat(img);
   glTexImage2D(GL_TEXTURE_2D,0,4,rgba.width(),rgba.height(),0,GL_RGBA,GL_UNSIGNED_BYTE,rgba.bits());
   //  Set pixel interpolation
   glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
   glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
   return tex;
}

//
//  Initialize
//
void Ex15opengl::initializeGL()
{
   if (init) return;
   init = true;

   //  Load shaders
   Shader(shader,":/ex15.vert",":/ex15.frag");

   //  Load textures
   day[0]  = LoadImage("day01.png");
   day[1]  = LoadImage("day02.png");
   day[2]  = LoadImage("day03.png");
   day[3]  = LoadImage("day04.png");
   day[4]  = LoadImage("day05.png");
   day[5]  = LoadImage("day06.png");
   day[6]  = LoadImage("day07.png");
   day[7]  = LoadImage("day08.png");
   day[8]  = LoadImage("day09.png");
   day[9]  = LoadImage("day10.png");
   day[10] = LoadImage("day11.png");
   day[11] = LoadImage("day12.png");
   night   = LoadImage("night.png");
   cge     = LoadImage("cge.png");

   //  Start 100 fps timer connected to updateGL
   timer.setInterval(10);
   connect(&timer,SIGNAL(timeout()),this,SLOT(updateGL()));
   timer.start();
   time.start();
}

//
//  Set projection when window is resized
//
void Ex15opengl::resizeGL(int width, int height)
{
   //  Window aspect ration
   asp = height ? width / (float)height : 1;
   //  Viewport is whole screen
   glViewport(0,0,width,height);
}

//
//  Draw the window
//
void Ex15opengl::paintGL()
{
   //  Set projection
   glMatrixMode(GL_PROJECTION);
   glLoadIdentity();
   glOrtho(-dim*asp, +dim*asp, -dim, +dim, -dim, +dim);
   //  Set view
   glMatrixMode(GL_MODELVIEW);
   glLoadIdentity();
   glRotated(th,0,1,0);
   glRotated(ph,1,0,0);
   glRotatef(-90,1,0,0);

   //  Wall time (15 second increments)
   float t = 0.001*time.elapsed()/15;
   float f = t-int(t);
   float zh = mode ? 360*f : 360*pct;

   //  Time of year
   float foy = mode ? pct : f;

   //  Compute month and day
   int dy = 365*fmod(foy,1)+1;
   int ndm[] = {31,28,31,30,31,30,31,31,30,31,30,31};
   int mo = 0;
   while (dy>ndm[mo])
      dy -= ndm[mo++];
   float frac = (dy-1)/(float)ndm[mo];
   mo++;
   char date[8];
   sprintf(date,"%.2d/%.2d",mo,dy);
   emit datim(QString(date));

   //  Light color and direction
   float Ambient[]  = {0.3,0.3,0.3,1.0};
   float Diffuse[]  = {0.8,0.8,0.8,1.0};
   float Specular[] = {1.0,1.0,1.0,1.0};
   float Position[]  = {0.0,-1.0,(float)(-0.43*Cos(360*foy)),0.0};

   //  Z-buffer
   glEnable(GL_DEPTH_TEST);
   //  Clear screen and depth buffer
   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
   //  OpenGL should normalize normal vectors
   glEnable(GL_NORMALIZE);
   //  Enable lighting
   glEnable(GL_LIGHTING);
   //  Enable light 0
   glEnable(GL_LIGHT0);
   //  Set ambient, diffuse, specular components and position of light 0
   glLightfv(GL_LIGHT0,GL_AMBIENT ,Ambient);
   glLightfv(GL_LIGHT0,GL_DIFFUSE ,Diffuse);
   glLightfv(GL_LIGHT0,GL_SPECULAR,Specular);
   glLightfv(GL_LIGHT0,GL_POSITION,Position);

   //  Texture for this month
   glActiveTexture(GL_TEXTURE0);
   glBindTexture(GL_TEXTURE_2D,day[mo-1]);
   //  Texture for next month
   glActiveTexture(GL_TEXTURE1);
   glBindTexture(GL_TEXTURE_2D,day[mo%12]);
   //  Load nightime texture
   glActiveTexture(GL_TEXTURE2);
   glBindTexture(GL_TEXTURE_2D,night);
   //  Load cloud, gloss and elevation texture
   glActiveTexture(GL_TEXTURE3);
   glBindTexture(GL_TEXTURE_2D,cge);

   //  Enable shader
   shader.bind();

   //  Set shader increments
   shader.setUniformValue("mode",mode);
   shader.setUniformValue("frac",frac);
   shader.setUniformValue("DayTex0" ,0);
   shader.setUniformValue("DayTex1" ,1);
   shader.setUniformValue("NightTex",2);
   shader.setUniformValue("CGE"     ,3);

   //  Draw earth
   glColor3f(1,1,1);
   glRotatef(zh,0,0,1);
   ball(0,0,0,1);

   //  Enable shader
   shader.release();

   //  Disable lighting
   glDisable(GL_LIGHTING);
   //  Draw axis
   glColor3f(1,1,1);
   glBegin(GL_LINES);
   glVertex3f(0,0,+1.2);
   glVertex3f(0,0,-1.2);
   glEnd();

   //  Emit date and time to display
   emit angles(QString::number(th)+","+QString::number(ph));
}

//
//  Throw a fatal error and die
//
void Ex15opengl::Fatal(QString message)
{
   QMessageBox::critical(this,"Ex15opengl",message);
   QApplication::quit();
}

/******************************************************************/
/*************************  Mouse Events  *************************/
/******************************************************************/
//
//  Mouse pressed
//
void Ex15opengl::mousePressEvent(QMouseEvent* e)
{
   mouse = true;
   pos = e->pos();  //  Remember mouse location
}

//
//  Mouse released
//
void Ex15opengl::mouseReleaseEvent(QMouseEvent*)
{
    mouse = false;
}

//
//  Mouse moved
//
void Ex15opengl::mouseMoveEvent(QMouseEvent* e)
{
   if (mouse)
   {
      QPoint d = e->pos()-pos;  //  Change in mouse location
      th = (th+d.x())%360;      //  Translate x movement to azimuth
      ph = (ph+d.y())%360;      //  Translate y movement to elevation
      pos = e->pos();           //  Remember new location
      updateGL();               //  Request redisplay
   }
}

//
//  Mouse wheel
//
void Ex15opengl::wheelEvent(QWheelEvent* e)
{
   //  Zoom out
   if (e->delta()<0)
      dim += 0.1;
   //  Zoom in
   else if (dim>1)
      dim -= 0.1;
   //  Request redisplay
   updateGL();
}

//
//  Load shader
//
void Ex15opengl::Shader(QGLShaderProgram& shader,QString vert,QString frag)
{
   //  Vertex shader
   if (vert.length() && !shader.addShaderFromSourceFile(QGLShader::Vertex,vert))
      Fatal("Error compiling "+vert+"\n"+shader.log());
   //  Fragment shader
   if (frag.length() && !shader.addShaderFromSourceFile(QGLShader::Fragment,frag))
      Fatal("Error compiling "+frag+"\n"+shader.log());
   //  Link
   if (!shader.link())
      Fatal("Error linking shader\n"+shader.log());
}
